# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from conftest import async_handler_support, skip_if_not_async_handler_support
from testing_support.fixtures import (
    override_application_settings,
    validate_tt_parenting,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.packages import six

try:
    # The __version__ attribute was only added in 0.7.0.
    # Flask team does not use semantic versioning during development.
    from flask import __version__ as flask_version

    flask_version = tuple([int(v) for v in flask_version.split(".")])
    is_gt_flask060 = True
    is_dev_version = False
except ValueError:
    is_gt_flask060 = True
    is_dev_version = True
except ImportError:
    is_gt_flask060 = False
    is_dev_version = False

requires_endpoint_decorator = pytest.mark.skipif(not is_gt_flask060, reason="The endpoint decorator is not supported.")


def target_application():
    # We need to delay Flask application creation because of ordering
    # issues whereby the agent needs to be initialised before Flask is
    # imported and the routes configured. Normally pytest only runs the
    # global fixture which will initialise the agent after each test
    # file is imported, which is too late. We also can't do application
    # creation within a function as we will then get view handler
    # functions are different between Python 2 and 3, with the latter
    # showing <local> scope in path.

    if not async_handler_support:
        from _test_application import _test_application
    else:
        from _test_application_async import _test_application
    return _test_application


_test_application_index_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_test_application:index_page", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
]

_test_application_index_tt_parenting = (
    "TransactionNode",
    [
        (
            "FunctionNode",
            [
                (
                    "FunctionNode",
                    [
                        ("FunctionNode", []),
                        ("FunctionNode", []),
                        ("FunctionNode", []),
                        # some flask versions have more FunctionNodes here, as appended
                        # below
                    ],
                ),
            ],
        ),
        ("FunctionNode", []),
        (
            "FunctionNode",
            [
                ("FunctionNode", []),
            ],
        ),
    ],
)

if is_dev_version or (is_gt_flask060 and flask_version >= (0, 7)):
    _test_application_index_tt_parenting[1][0][1][0][1].append(
        ("FunctionNode", []),
    )
if is_dev_version or (is_gt_flask060 and flask_version >= (0, 9)):
    _test_application_index_tt_parenting[1][0][1][0][1].append(
        ("FunctionNode", []),
    )


@validate_transaction_errors(errors=[])
@validate_transaction_metrics("_test_application:index_page", scoped_metrics=_test_application_index_scoped_metrics)
@validate_tt_parenting(_test_application_index_tt_parenting)
@validate_code_level_metrics("_test_application", "index_page")
def test_application_index():
    application = target_application()
    response = application.get("/index")
    response.mustcontain("INDEX RESPONSE")


_test_application_async_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_test_application_async:async_page", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
]


@skip_if_not_async_handler_support
@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_test_application_async:async_page", scoped_metrics=_test_application_async_scoped_metrics
)
@validate_tt_parenting(_test_application_index_tt_parenting)
@validate_code_level_metrics("_test_application_async", "async_page")
def test_application_async():
    application = target_application()
    response = application.get("/async")
    response.mustcontain("ASYNC RESPONSE")


_test_application_endpoint_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_test_application:endpoint_page", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_test_application:endpoint_page", scoped_metrics=_test_application_endpoint_scoped_metrics
)
@validate_code_level_metrics("_test_application", "endpoint_page")
def test_application_endpoint():
    application = target_application()
    response = application.get("/endpoint")
    response.mustcontain("ENDPOINT RESPONSE")


_test_application_error_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_test_application:error_page", 1),
    ("Function/flask.app:Flask.handle_exception", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
    ("Function/flask.app:Flask.handle_user_exception", 1),
    ("Function/flask.app:Flask.handle_user_exception", 1),
]


if six.PY3:
    _test_application_error_errors = ["builtins:RuntimeError"]
else:
    _test_application_error_errors = ["exceptions:RuntimeError"]


@validate_transaction_errors(errors=_test_application_error_errors)
@validate_transaction_metrics("_test_application:error_page", scoped_metrics=_test_application_error_scoped_metrics)
@validate_code_level_metrics("_test_application", "error_page")
def test_application_error():
    application = target_application()
    application.get("/error", status=500, expect_errors=True)


_test_application_abort_404_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_test_application:abort_404_page", 1),
    ("Function/flask.app:Flask.handle_http_exception", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
    ("Function/flask.app:Flask.handle_user_exception", 1),
]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_test_application:abort_404_page", scoped_metrics=_test_application_abort_404_scoped_metrics
)
@validate_code_level_metrics("_test_application", "abort_404_page")
def test_application_abort_404():
    application = target_application()
    application.get("/abort_404", status=404)


_test_application_exception_404_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_test_application:exception_404_page", 1),
    ("Function/flask.app:Flask.handle_http_exception", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
    ("Function/flask.app:Flask.handle_user_exception", 1),
]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_test_application:exception_404_page", scoped_metrics=_test_application_exception_404_scoped_metrics
)
@validate_code_level_metrics("_test_application", "exception_404_page")
def test_application_exception_404():
    application = target_application()
    application.get("/exception_404", status=404)


_test_application_not_found_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/flask.app:Flask.handle_http_exception", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
    ("Function/flask.app:Flask.handle_user_exception", 1),
]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "flask.app:Flask.handle_http_exception", scoped_metrics=_test_application_not_found_scoped_metrics
)
def test_application_not_found():
    application = target_application()
    application.get("/missing", status=404)


_test_application_render_template_string_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_test_application:template_string", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
    ("Template/Compile/<template>", 1),
    ("Template/Render/<template>", 1),
]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_test_application:template_string", scoped_metrics=_test_application_render_template_string_scoped_metrics
)
@validate_code_level_metrics("_test_application", "template_string")
def test_application_render_template_string():
    application = target_application()
    application.get("/template_string")


_test_application_render_template_not_found_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_test_application:template_not_found", 1),
    ("Function/flask.app:Flask.handle_exception", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
    ("Function/flask.app:Flask.handle_user_exception", 1),
]


@validate_transaction_errors(errors=["jinja2.exceptions:TemplateNotFound"])
@validate_transaction_metrics(
    "_test_application:template_not_found", scoped_metrics=_test_application_render_template_not_found_scoped_metrics
)
def test_application_render_template_not_found():
    application = target_application()
    application.get("/template_not_found", status=500, expect_errors=True)


_test_html_insertion_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion():
    application = target_application()
    response = application.get("/html_insertion", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain("NREUM HEADER", "NREUM.info")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from testing_support.fixtures import catch_background_exceptions

from newrelic.common.object_wrapper import function_wrapper, transient_function_wrapper


def validate_ml_event_count_outside_transaction(count=1):
    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):

        record_called = []
        recorded_events = []

        @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_ml_event")
        @catch_background_exceptions
        def _validate_ml_event_count_outside_transaction(wrapped, instance, args, kwargs):
            record_called.append(True)
            try:
                result = wrapped(*args, **kwargs)
            except:
                raise
            recorded_events[:] = []
            recorded_events.extend(list(instance._ml_events))

            return result

        _new_wrapper = _validate_ml_event_count_outside_transaction(wrapped)
        val = _new_wrapper(*args, **kwargs)
        if count:
            assert record_called
        events = copy.copy(recorded_events)

        record_called[:] = []
        recorded_events[:] = []

        assert count == len(events), len(events)

        return val

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import os

import pika
import pytest
import six
from compat import basic_consume
from conftest import BODY, CORRELATION_ID, EXCHANGE, HEADERS, QUEUE, REPLY_TO
from testing_support.db_settings import rabbitmq_settings
from testing_support.fixtures import capture_transaction_metrics
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task
from newrelic.api.transaction import end_of_transaction

DB_SETTINGS = rabbitmq_settings()[0]

_message_broker_tt_params = {
    'queue_name': QUEUE,
    'routing_key': QUEUE,
    'correlation_id': CORRELATION_ID,
    'reply_to': REPLY_TO,
    'headers': HEADERS.copy(),
}

_test_blocking_connection_basic_get_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE, 1),
    (('Function/pika.adapters.blocking_connection:'
            '_CallbackResult.set_value_once'), 1)
]


@validate_transaction_metrics(
        ('test_pika_blocking_connection_consume:'
                'test_blocking_connection_basic_get'),
        scoped_metrics=_test_blocking_connection_basic_get_metrics,
        rollup_metrics=_test_blocking_connection_basic_get_metrics,
        background_task=True)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_basic_get(producer):
    with pika.BlockingConnection(
            pika.ConnectionParameters(DB_SETTINGS['host'])) as connection:
        channel = connection.channel()
        method_frame, _, _ = channel.basic_get(QUEUE)
        assert method_frame
        channel.basic_ack(method_frame.delivery_tag)


_test_blocking_connection_basic_get_empty_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE, None),
]


@validate_transaction_metrics(
        ('test_pika_blocking_connection_consume:'
                'test_blocking_connection_basic_get_empty'),
        scoped_metrics=_test_blocking_connection_basic_get_empty_metrics,
        rollup_metrics=_test_blocking_connection_basic_get_empty_metrics,
        background_task=True)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_basic_get_empty():
    QUEUE = 'test_blocking_empty-%s' % os.getpid()
    with pika.BlockingConnection(
            pika.ConnectionParameters(DB_SETTINGS['host'])) as connection:
        channel = connection.channel()
        channel.queue_declare(queue=QUEUE)

        try:
            method_frame, _, _ = channel.basic_get(QUEUE)
            assert method_frame is None
        finally:
            channel.queue_delete(queue=QUEUE)


def test_blocking_connection_basic_get_outside_transaction(producer):
    metrics_list = []

    @capture_transaction_metrics(metrics_list)
    def test_basic_get():
        with pika.BlockingConnection(
                pika.ConnectionParameters(DB_SETTINGS['host'])) as connection:
            channel = connection.channel()
            channel.queue_declare(queue=QUEUE)

            method_frame, _, _ = channel.basic_get(QUEUE)
            channel.basic_ack(method_frame.delivery_tag)
            assert method_frame

    test_basic_get()

    # Confirm that no metrics have been created. This is because no background
    # task should be created for basic_get actions.
    assert not metrics_list


_test_blocking_conn_basic_consume_no_txn_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE, None),
]

if six.PY3:
    _txn_name = ('test_pika_blocking_connection_consume:'
            'test_blocking_connection_basic_consume_outside_transaction.'
            '<locals>.on_message')
    _test_blocking_conn_basic_consume_no_txn_metrics.append(
        (('Function/test_pika_blocking_connection_consume:'
          'test_blocking_connection_basic_consume_outside_transaction.'
          '<locals>.on_message'), None))
else:
    _txn_name = ('test_pika_blocking_connection_consume:'
            'on_message')
    _test_blocking_conn_basic_consume_no_txn_metrics.append(
        ('Function/test_pika_blocking_connection_consume:on_message', None))


@pytest.mark.parametrize('as_partial', [True, False])
@validate_code_level_metrics("test_pika_blocking_connection_consume.test_blocking_connection_basic_consume_outside_transaction.<locals>", "on_message", py2_namespace="test_pika_blocking_connection_consume")
@validate_transaction_metrics(
        _txn_name,
        scoped_metrics=_test_blocking_conn_basic_consume_no_txn_metrics,
        rollup_metrics=_test_blocking_conn_basic_consume_no_txn_metrics,
        background_task=True,
        group='Message/RabbitMQ/Exchange/%s' % EXCHANGE)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
def test_blocking_connection_basic_consume_outside_transaction(producer,
        as_partial):
    def on_message(channel, method_frame, header_frame, body):
        assert hasattr(method_frame, '_nr_start_time')
        assert body == BODY
        channel.stop_consuming()

    if as_partial:
        on_message = functools.partial(on_message)

    with pika.BlockingConnection(
            pika.ConnectionParameters(DB_SETTINGS['host'])) as connection:
        channel = connection.channel()

        basic_consume(channel, QUEUE, on_message)
        try:
            channel.start_consuming()
        except:
            channel.stop_consuming()
            raise


_test_blocking_conn_basic_consume_in_txn_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE, None),
]

if six.PY3:
    _test_blocking_conn_basic_consume_in_txn_metrics.append(
        (('Function/test_pika_blocking_connection_consume:'
          'test_blocking_connection_basic_consume_inside_txn.'
          '<locals>.on_message'), 1))
else:
    _test_blocking_conn_basic_consume_in_txn_metrics.append(
        ('Function/test_pika_blocking_connection_consume:on_message', 1))


@pytest.mark.parametrize('as_partial', [True, False])
@validate_code_level_metrics("test_pika_blocking_connection_consume.test_blocking_connection_basic_consume_inside_txn.<locals>", "on_message", py2_namespace="test_pika_blocking_connection_consume")
@validate_transaction_metrics(
        ('test_pika_blocking_connection_consume:'
                'test_blocking_connection_basic_consume_inside_txn'),
        scoped_metrics=_test_blocking_conn_basic_consume_in_txn_metrics,
        rollup_metrics=_test_blocking_conn_basic_consume_in_txn_metrics,
        background_task=True)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_basic_consume_inside_txn(producer, as_partial):
    def on_message(channel, method_frame, header_frame, body):
        assert hasattr(method_frame, '_nr_start_time')
        assert body == BODY
        channel.stop_consuming()

    if as_partial:
        on_message = functools.partial(on_message)

    with pika.BlockingConnection(
            pika.ConnectionParameters(DB_SETTINGS['host'])) as connection:
        channel = connection.channel()
        basic_consume(channel, QUEUE, on_message)
        try:
            channel.start_consuming()
        except:
            channel.stop_consuming()
            raise


_test_blocking_conn_basic_consume_stopped_txn_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE, None),
    ('OtherTransaction/Message/RabbitMQ/Exchange/Named/%s' % EXCHANGE, None),
]

if six.PY3:
    _test_blocking_conn_basic_consume_stopped_txn_metrics.append(
        (('Function/test_pika_blocking_connection_consume:'
          'test_blocking_connection_basic_consume_stopped_txn.'
          '<locals>.on_message'), None))
else:
    _test_blocking_conn_basic_consume_stopped_txn_metrics.append(
        ('Function/test_pika_blocking_connection_consume:on_message', None))


@pytest.mark.parametrize('as_partial', [True, False])
@validate_transaction_metrics(
        ('test_pika_blocking_connection_consume:'
                'test_blocking_connection_basic_consume_stopped_txn'),
        scoped_metrics=_test_blocking_conn_basic_consume_stopped_txn_metrics,
        rollup_metrics=_test_blocking_conn_basic_consume_stopped_txn_metrics,
        background_task=True)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_blocking_connection_basic_consume_stopped_txn(producer, as_partial):
    def on_message(channel, method_frame, header_frame, body):
        assert hasattr(method_frame, '_nr_start_time')
        assert body == BODY
        channel.stop_consuming()

    end_of_transaction()

    if as_partial:
        on_message = functools.partial(on_message)

    with pika.BlockingConnection(
            pika.ConnectionParameters(DB_SETTINGS['host'])) as connection:
        channel = connection.channel()
        basic_consume(channel, QUEUE, on_message)
        try:
            channel.start_consuming()
        except:
            channel.stop_consuming()
            raise



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from graphene import Field, Int, List
from graphene import Mutation as GrapheneMutation
from graphene import NonNull, ObjectType, Schema, String, Union


class Author(ObjectType):
    first_name = String()
    last_name = String()


class Book(ObjectType):
    id = Int()
    name = String()
    isbn = String()
    author = Field(Author)
    branch = String()


class Magazine(ObjectType):
    id = Int()
    name = String()
    issue = Int()
    branch = String()


class Item(Union):
    class Meta:
        types = (Book, Magazine)


class Library(ObjectType):
    id = Int()
    branch = String()
    magazine = Field(List(Magazine))
    book = Field(List(Book))


Storage = List(String)


authors = [
    Author(
        first_name="New",
        last_name="Relic",
    ),
    Author(
        first_name="Bob",
        last_name="Smith",
    ),
    Author(
        first_name="Leslie",
        last_name="Jones",
    ),
]

books = [
    Book(
        id=1,
        name="Python Agent: The Book",
        isbn="a-fake-isbn",
        author=authors[0],
        branch="riverside",
    ),
    Book(
        id=2,
        name="Ollies for O11y: A Sk8er's Guide to Observability",
        isbn="a-second-fake-isbn",
        author=authors[1],
        branch="downtown",
    ),
    Book(
        id=3,
        name="[Redacted]",
        isbn="a-third-fake-isbn",
        author=authors[2],
        branch="riverside",
    ),
]

magazines = [
    Magazine(id=1, name="Reli Updates Weekly", issue=1, branch="riverside"),
    Magazine(id=2, name="Reli Updates Weekly", issue=2, branch="downtown"),
    Magazine(id=3, name="Node Weekly", issue=1, branch="riverside"),
]


libraries = ["riverside", "downtown"]
libraries = [
    Library(
        id=i + 1,
        branch=branch,
        magazine=[m for m in magazines if m.branch == branch],
        book=[b for b in books if b.branch == branch],
    )
    for i, branch in enumerate(libraries)
]

storage = []


def resolve_library(self, info, index):
    return libraries[index]

def resolve_storage(self, info):
    return [storage.pop()]

def resolve_search(self, info, contains):
    search_books = [b for b in books if contains in b.name]
    search_magazines = [m for m in magazines if contains in m.name]
    return search_books + search_magazines

def resolve_hello(self, info):
    return "Hello!"

def resolve_echo(self, info, echo):
    return echo

def resolve_error(self, info):
    raise RuntimeError("Runtime Error!")

def resolve_storage_add(self, info, string):
    storage.append(string)
    return StorageAdd(string=string)


class StorageAdd(GrapheneMutation):
    class Arguments:
        string = String(required=True)

    string = String()
    mutate = resolve_storage_add


class Query(ObjectType):
    library = Field(Library, index=Int(required=True), resolver=resolve_library)
    hello = String(resolver=resolve_hello)
    search = Field(List(Item), contains=String(required=True), resolver=resolve_search)
    echo = Field(String, echo=String(required=True), resolver=resolve_echo)
    storage = Field(Storage, resolver=resolve_storage)
    error = String(resolver=resolve_error)
    error_non_null = Field(NonNull(String), resolver=resolve_error)
    error_middleware = String(resolver=resolve_hello)


class Mutation(ObjectType):
    storage_add = StorageAdd.Field()


target_schema = Schema(query=Query, mutation=Mutation, auto_camelcase=False)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import time

import pytest
from conftest import cache_kafka_producer_headers
from testing_support.validators.validate_messagebroker_headers import (
    validate_messagebroker_headers,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.api.function_trace import FunctionTrace
from newrelic.common.object_names import callable_name
from newrelic.packages import six


@pytest.mark.parametrize(
    "headers", [[("MY-HEADER", "nonsense")], {"MY-HEADER": "nonsense"}], ids=["list of tuples headers", "dict headers"]
)
@background_task()
def test_produce_arguments(topic, producer, client_type, serialize, headers):
    callback1_called = threading.Event()
    callback2_called = threading.Event()
    ts = int(time.time())

    def producer_callback1(err, msg):
        callback1_called.set()

    def producer_callback2(err, msg):
        callback2_called.set()

    if client_type == "cimpl":
        # Keyword Args
        producer.produce(
            topic=topic,
            value=serialize({"foo": 1}),
            key=serialize("my-key"),
            partition=0,
            callback=producer_callback2,
            timestamp=ts,
            headers=headers,
        )
        # Positional Args
        producer.produce(
            topic,
            serialize({"foo": 1}),
            serialize("my-key"),
            0,
            producer_callback1,
            None,
            ts,
            headers,
        )
    else:
        # Keyword Args
        producer.produce(
            topic=topic,
            value=serialize({"foo": 1}),
            key=serialize("my-key"),
            partition=0,
            on_delivery=producer_callback2,
            timestamp=ts,
            headers=headers,
        )
        # Positional Args
        producer.produce(
            topic,
            serialize("my-key"),
            serialize({"foo": 1}),
            0,
            producer_callback1,
            ts,
            headers,
        )
    producer.flush()

    assert callback1_called.wait(5), "Callback never called."
    assert callback2_called.wait(5), "Callback never called."


def test_trace_metrics(topic, send_producer_message):
    from confluent_kafka import __version__ as version

    scoped_metrics = [("MessageBroker/Kafka/Topic/Produce/Named/%s" % topic, 1)]
    unscoped_metrics = scoped_metrics
    txn_name = "test_producer:test_trace_metrics.<locals>.test" if six.PY3 else "test_producer:test"

    @validate_transaction_metrics(
        txn_name,
        scoped_metrics=scoped_metrics,
        rollup_metrics=unscoped_metrics,
        custom_metrics=[("Python/MessageBroker/Confluent-Kafka/%s" % version, 1)],
        background_task=True,
    )
    @background_task()
    def test():
        send_producer_message()

    test()


def test_distributed_tracing_headers(topic, send_producer_message):
    txn_name = "test_producer:test_distributed_tracing_headers.<locals>.test" if six.PY3 else "test_producer:test"

    @validate_transaction_metrics(
        txn_name,
        rollup_metrics=[
            ("Supportability/TraceContext/Create/Success", 1),
            ("Supportability/DistributedTrace/CreatePayload/Success", 1),
        ],
        background_task=True,
    )
    @background_task()
    @cache_kafka_producer_headers()
    @validate_messagebroker_headers
    def test():
        send_producer_message()

    test()


def test_distributed_tracing_headers_under_terminal(topic, send_producer_message):
    @validate_transaction_metrics(
        "test_distributed_tracing_headers_under_terminal",
        rollup_metrics=[
            ("Supportability/TraceContext/Create/Success", 1),
            ("Supportability/DistributedTrace/CreatePayload/Success", 1),
        ],
        background_task=True,
    )
    @background_task(name="test_distributed_tracing_headers_under_terminal")
    @cache_kafka_producer_headers()
    @validate_messagebroker_headers
    def test():
        with FunctionTrace(name="terminal_trace", terminal=True):
            send_producer_message()

    test()


def test_producer_errors(topic, producer, monkeypatch):
    if hasattr(producer, "_value_serializer"):
        # Remove serializer to intentionally cause a type error in underlying producer implementation
        monkeypatch.setattr(producer, "_value_serializer", None)

    @validate_transaction_errors([callable_name(TypeError)])
    @background_task()
    def test():
        with pytest.raises(TypeError):
            producer.produce(topic, value=object())
            producer.flush()

    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import FunctionWrapper, wrap_function_wrapper

# This is NOT a fully-featured instrumentation for the sentry SDK. Instead
# this is a monkey-patch of the sentry SDK to work around a bug that causes
# improper ASGI 2/3 version detection when inspecting our wrappers. We fix this
# by manually unwrapping the application when version detection is run.


def bind__looks_like_asgi3(app):
    return app


def wrap__looks_like_asgi3(wrapped, instance, args, kwargs):
    try:
        app = bind__looks_like_asgi3(*args, **kwargs)
    except Exception:
        return wrapped(*args, **kwargs)

    while isinstance(app, FunctionWrapper) and hasattr(app, "__wrapped__"):
        app = app.__wrapped__

    return wrapped(app)


def instrument_sentry_sdk_integrations_asgi(module):
    if hasattr(module, "_looks_like_asgi3"):
        wrap_function_wrapper(module, "_looks_like_asgi3", wrap__looks_like_asgi3)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from newrelic.hooks.database_psycopg2 import (instance_info, _add_defaults,
        _parse_connect_params)

def test_kwargs():
    connect_params = ((), {'database': 'foo', 'host': '1.2.3.4', 'port': 1234})
    output = _parse_connect_params(*connect_params)
    assert output == ('1.2.3.4', None, '1234', 'foo')

def test_arg_str():
    connect_params = (("host=foobar port=9876",), {})
    output = _parse_connect_params(*connect_params)
    assert output == ('foobar', None, '9876', None)

def test_bind_dsn():
    connect_params = ((), {'dsn': 'host=foobar port=9876'})
    output = _parse_connect_params(*connect_params)
    assert output == ('foobar', None, '9876', None)

def test_bind_dsn_ignore_kwargs():
    connect_params = ((), {'dsn': "host=foobar", 'port': 1234})
    output = _parse_connect_params(*connect_params)
    assert output == ('foobar', None, None, None)

def test_kwargs_str_for_port():
    connect_params = ((), {'database': 'foo', 'host': '1.2.3.4', 'port': '1234'})
    output = _parse_connect_params(*connect_params)
    assert output == ('1.2.3.4', None, '1234', 'foo')

def test_arg_str_missing_port():
    connect_params = (("host=foobar",), {})
    output = _parse_connect_params(*connect_params)
    assert output == ('foobar', None, None, None)

def test_arg_str_multiple_host():
    connect_params = (("host=foobar host=barbaz",), {})
    output = _parse_connect_params(*connect_params)
    assert output == ('barbaz', None, None, None)

def test_arg_str_multiple_port():
    connect_params = (("port=5555 port=7777",), {})
    output = _parse_connect_params(*connect_params)
    assert output == (None, None, '7777', None)

def test_arg_str_missing_host():
    connect_params = (("port=5555",), {})
    output = _parse_connect_params(*connect_params)
    assert output == (None, None, '5555', None)

def test_arg_str_missing_host_and_port():
    connect_params = (("nothing=here",), {})
    output = _parse_connect_params(*connect_params)
    assert output == (None, None, None, None)

def test_malformed_arg_str():
    connect_params = (("this_is_malformed",), {})
    output = _parse_connect_params(*connect_params)
    assert output == ('unknown', 'unknown', 'unknown', 'unknown')

def test_str_in_port_arg_str():
    connect_params = (("port=foobar",), {})
    output = _parse_connect_params(*connect_params)
    assert output == (None, None, 'foobar', None)

def test_host_and_hostaddr_in_arg_str():
    connect_params = (("host=foobar hostaddr=1.2.3.4",), {})
    output = _parse_connect_params(*connect_params)
    assert output == ('foobar', '1.2.3.4', None, None)

def test_host_and_hostaddr_in_kwarg():
    connect_params = ((), {'host':'foobar', 'hostaddr':'1.2.3.4'})
    output = _parse_connect_params(*connect_params)
    assert output == ('foobar', '1.2.3.4', None, None)

def test_only_hostaddr_in_kwarg():
    connect_params = ((), {'hostaddr':'1.2.3.4'})
    output = _parse_connect_params(*connect_params)
    assert output == (None, '1.2.3.4', None, None)

def test_only_hostaddr_in_arg_str():
    connect_params = (("hostaddr=1.2.3.4",), {})
    output = _parse_connect_params(*connect_params)
    assert output == (None, '1.2.3.4', None, None)

def test_env_var_default_host(monkeypatch):
    monkeypatch.setenv('PGHOST', 'envfoo')
    output = _add_defaults(None, None, '1234', 'foo')
    assert output == ('envfoo', '1234', 'foo')

def test_env_var_default_hostaddr(monkeypatch):
    monkeypatch.setenv('PGHOSTADDR', '1.2.3.4')
    output = _add_defaults(None, None, '1234', 'foo')
    assert output == ('1.2.3.4', '1234', 'foo')

def test_env_var_default_database(monkeypatch):
    monkeypatch.setenv('PGDATABASE', 'dbenvfoo')
    output = _add_defaults('foo', None, '1234', None)
    assert output == ('foo', '1234', 'dbenvfoo')

def test_env_var_default_port(monkeypatch):
    monkeypatch.setenv('PGPORT', '9999')
    output = _add_defaults('foo', None, None, 'bar')
    assert output == ('foo', '9999', 'bar')

@pytest.mark.parametrize('connect_params,expected', [
    ((('postgresql://',), {}),
        ('localhost', 'default', 'default')),
    ((('postgresql://localhost',), {}),
        ('localhost', '5432', 'default')),
    ((('postgresql://localhost:5433',), {}),
        ('localhost', '5433', 'default')),
    ((('postgresql://localhost/mydb',), {}),
        ('localhost', '5432', 'mydb')),
    ((('postgresql://user@localhost',), {}),
        ('localhost', '5432', 'default')),
    ((('postgresql://user:secret@localhost',), {}),
        ('localhost', '5432', 'default')),
    ((('postgresql://[2001:db8::1234]/database',), {}),
        ('2001:db8::1234', '5432', 'database')),
    ((('postgresql://[2001:db8::1234]:2222/database',), {}),
        ('2001:db8::1234', '2222', 'database')),
    ((('postgresql:///dbname?host=/var/lib/postgresql',), {}),
        ('localhost', '/var/lib/postgresql/.s.PGSQL.5432', 'dbname')),
    ((('postgresql://%2Fvar%2Flib%2Fpostgresql/dbname',), {}),
        ('localhost', '/var/lib/postgresql/.s.PGSQL.5432', 'dbname')),
    ((('postgresql://other@localhost/otherdb?c=10&a=myapp',), {}),
        ('localhost', '5432', 'otherdb')),
    ((('postgresql:///',), {}),
        ('localhost', 'default', 'default')),
    ((('postgresql:///dbname?host=foo',), {}),
        ('foo', '5432', 'dbname')),
    ((('postgresql:///dbname?port=1234',), {}),
        ('localhost', 'default', 'dbname')),
    ((('postgresql:///dbname?host=foo&port=1234',), {}),
        ('foo', '1234', 'dbname')),
    ((('postgres:///dbname?host=foo&port=1234',), {}),
        ('foo', '1234', 'dbname')),
    ((('postgres://localhost:5444/blah?host=::1',), {}),
        ('::1', '5444', 'blah')),
    ((('postgresql:///dbname?host=foo&port=1234&host=bar',), {}),
        ('bar', '1234', 'dbname')),
    ((('postgresql://%2Ftmp:1234',), {}),
        ('localhost', '/tmp/.s.PGSQL.1234', 'default')),
    ((('postgresql:///foo?dbname=bar',), {}),
        ('localhost', 'default', 'bar')),
    ((('postgresql://example.com/foo?hostaddr=1.2.3.4&host=bar',), {}),
        ('1.2.3.4', '5432', 'foo')),
])
def test_uri(connect_params, expected):
    output = instance_info(*connect_params)
    assert output == expected

@pytest.mark.parametrize('connect_params,expected', [
    ((('postgresql://user:password@/?dbname=bar',), {}),
        ('localhost', 'default', 'bar')),
    ((('postgresql://user:pass@host/?dbname=bar',), {}),
        ('host', '5432', 'bar')),
    ((('postgresql://user:password@@/?dbname=bar',), {}),
        ('localhost', 'default', 'bar')),
    ((('postgresql://@',), {}),
        ('localhost', 'default', 'default')),
    ((('postgresql://@@localhost',), {}),
        ('localhost', '5432', 'default')),
])
def test_security_sensitive_uri(connect_params, expected):
    output = instance_info(*connect_params)
    assert output == expected

def test_bad_uri():
    connect_params = (("blah:///foo",), {})
    output = instance_info(*connect_params)
    assert output == ('unknown', 'unknown', 'unknown')

_test_add_defaults = [

    # TCP/IP

    [('otherhost.com', None, '8888', 'foobar'), ('otherhost.com', '8888', 'foobar')],
    [('otherhost.com', None, None, 'foobar'), ('otherhost.com', '5432', 'foobar')],
    [('localhost', None, '8888', 'foobar'), ('localhost', '8888', 'foobar')],
    [('localhost', None, None, 'foobar'), ('localhost', '5432', 'foobar')],
    [('127.0.0.1', None, '8888', 'foobar'), ('127.0.0.1', '8888', 'foobar')],
    [('127.0.0.1', None, None, 'foobar'), ('127.0.0.1', '5432', 'foobar')],
    [('::1', None, '8888', None), ('::1', '8888', 'default')],
    [('::1', None, None, None), ('::1', '5432', 'default')],
    [('::1', None, None, ''), ('::1', '5432', 'default')],

    # Unix Domain Socket

    [(None, None, None, None), ('localhost', 'default', 'default')],
    [(None, None, '5432', None), ('localhost', 'default', 'default')],
    [(None, None, '8888', None), ('localhost', 'default', 'default')],
    [('/tmp', None, None, 'cat'), ('localhost', '/tmp/.s.PGSQL.5432', 'cat')],
    [('/tmp', None, '5432', 'dog'), ('localhost', '/tmp/.s.PGSQL.5432', 'dog')],
    [('/tmp', None, '8888', 'db'), ('localhost', '/tmp/.s.PGSQL.8888', 'db')],
]

@pytest.mark.parametrize('host_port,expected', _test_add_defaults)
def test_add_defaults(host_port, expected):
    actual = _add_defaults(*host_port)
    assert actual == expected



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from newrelic.api.application import register_application
from newrelic.api.background_task import background_task
from testing_support.validators.validate_serverless_payload import (
        validate_serverless_payload)

INI_FILE_EMPTY = """
[newrelic]
""".encode('utf-8')

INI_FILE_SERVERLESS_MODE = """
[newrelic]
serverless_mode.enabled = true
""".encode('utf-8')

INI_FILE_APDEX_T = """
[newrelic]
apdex_t = 0.33
""".encode('utf-8')

INI_FILE_APP_NAME = """
[newrelic]
serverless_mode.enabled = True
app_name = a;b
""".encode('utf-8')

NON_SERVERLESS_MODE_ENV = {
    'NEW_RELIC_SERVERLESS_MODE_ENABLED': 'false',
}

SERVERLESS_MODE_ENV = {
    'NEW_RELIC_SERVERLESS_MODE_ENABLED': 'true',
}

DT_ENV = {
    'NEW_RELIC_ACCOUNT_ID': 'account_id',
    'NEW_RELIC_PRIMARY_APPLICATION_ID': 'application_id',
    'NEW_RELIC_TRUSTED_ACCOUNT_KEY': 'trusted_key',
}

APP_NAME_ENV = {
    'NEW_RELIC_APP_NAME': 'c;d'
}


@pytest.mark.parametrize('ini,env,serverless_mode', [
    # 1. serverless mode in config file (this trumps all)
    (INI_FILE_SERVERLESS_MODE, NON_SERVERLESS_MODE_ENV, True),

    # 2. if all else fails, NEW_RELIC_SERVERLESS_MODE_ENABLED should work
    (INI_FILE_EMPTY, SERVERLESS_MODE_ENV, True),

    # 3. Default is false
    (INI_FILE_EMPTY, {}, False),
])
def test_serverless_mode_environment(ini, env, serverless_mode,
        global_settings):
    settings = global_settings()
    assert settings.serverless_mode.enabled == serverless_mode


@pytest.mark.parametrize('ini,env,value', [
    # 1. apdex_t set in config file (this trumps all)
    (INI_FILE_APDEX_T, {'NEW_RELIC_APDEX_T': 0.25}, 0.33),

    # 2. lambda environment variable should force serverless mode on
    (INI_FILE_EMPTY, {'NEW_RELIC_APDEX_T': 0.25}, 0.25),

    # 3. Default is what's in config.py
    (INI_FILE_EMPTY, {}, 0.5),
])
def test_serverless_apdex_t(ini, env, value, global_settings):
    settings = global_settings()
    assert settings.apdex_t == value


@pytest.mark.parametrize('ini,env,values', [
    (INI_FILE_EMPTY, NON_SERVERLESS_MODE_ENV, (
            None, 'Unknown', None)),
    (INI_FILE_SERVERLESS_MODE, DT_ENV, (
            'account_id', 'application_id', 'trusted_key')),
])
def test_serverless_dt_environment(ini, env, values, global_settings):
    settings = global_settings()

    assert settings.account_id == values[0]
    assert settings.primary_application_id == values[1]
    assert settings.trusted_account_key == values[2]


# The purpose of test_serverless_app_names is to ensure that if the customer
# ever tries to specify multiple application names when serverless_mode = True,
# the agent will never output more than one payload.

@pytest.mark.parametrize('ini,env,register_name,handler_name', (
    # multiple app names specified via the newrelic.ini config file
    (INI_FILE_APP_NAME, {}, None, None),

    # multiple app names specified via the NEW_RELIC_APP_NAME env variable
    (INI_FILE_SERVERLESS_MODE, APP_NAME_ENV, None, None),

    # multiple app names specified via a string passed to register_application
    (INI_FILE_SERVERLESS_MODE, {}, 'e;f', None),

    # multiple app names specified via a string passed to transaction decorator
    (INI_FILE_SERVERLESS_MODE, {}, None, 'h;i'),
))
def test_serverless_mutli_app_names(ini, env, register_name, handler_name,
                                    global_settings):
    application = handler_name or register_application(register_name or None)

    @validate_serverless_payload()
    @background_task(application=application)
    def _test_serverless_app_names():
        pass

    _test_serverless_app_names()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pika
import pytest
from testing_support.db_settings import rabbitmq_settings
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_messagebroker_headers import (
    validate_messagebroker_headers,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import transient_function_wrapper


@transient_function_wrapper(pika.frame, "Header.__init__")
def cache_pika_headers(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    ret = wrapped(*args, **kwargs)
    headers = instance.properties.headers
    transaction._test_request_headers = headers
    return ret


DB_SETTINGS = rabbitmq_settings()[0]
QUEUE = "test-pika-queue"
CORRELATION_ID = "testingpika"
REPLY_TO = "testing"
HEADERS = {"MYHEADER": "pikatest"}

_message_broker_tt_included_params = {
    "routing_key": QUEUE,
}

_message_broker_tt_forgone_params = ["queue_name", "correlation_id", "reply_to"]

_test_blocking_connection_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/Default", 2),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/Default", None),
]


@validate_transaction_metrics(
    "test_pika_produce:test_blocking_connection",
    scoped_metrics=_test_blocking_connection_metrics,
    rollup_metrics=_test_blocking_connection_metrics,
    background_task=True,
)
@validate_tt_collector_json(
    message_broker_params=_message_broker_tt_included_params,
    message_broker_forgone_params=_message_broker_tt_forgone_params,
)
@background_task()
@validate_messagebroker_headers
@cache_pika_headers
def test_blocking_connection(producer):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        channel.basic_publish(
            exchange="",
            routing_key=QUEUE,
            body="test",
        )

        # publish has been removed and replaced with basic_publish in later
        # versions of pika
        publish = getattr(channel, "publish", channel.basic_publish)

        publish(
            exchange="",
            routing_key=QUEUE,
            body="test",
        )


_message_broker_tt_included_test_correlation_id = _message_broker_tt_included_params.copy()
_message_broker_tt_included_test_correlation_id.update(
    {
        "correlation_id": CORRELATION_ID,
    }
)

_message_broker_tt_forgone_test_correlation_id = ["queue_name", "reply_to", "headers"]


@validate_transaction_metrics(
    "test_pika_produce:test_blocking_connection_correlation_id",
    scoped_metrics=_test_blocking_connection_metrics,
    rollup_metrics=_test_blocking_connection_metrics,
    background_task=True,
)
@validate_tt_collector_json(
    message_broker_params=_message_broker_tt_included_test_correlation_id,
    message_broker_forgone_params=(_message_broker_tt_forgone_test_correlation_id),
)
@background_task()
@validate_messagebroker_headers
@cache_pika_headers
def test_blocking_connection_correlation_id(producer):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        channel.basic_publish(
            exchange="",
            routing_key=QUEUE,
            body="test",
            properties=pika.spec.BasicProperties(correlation_id=CORRELATION_ID),
        )

        # publish has been removed and replaced with basic_publish in later
        # versions of pika
        publish = getattr(channel, "publish", channel.basic_publish)

        publish(
            exchange="",
            routing_key=QUEUE,
            body="test",
            properties=pika.spec.BasicProperties(correlation_id=CORRELATION_ID),
        )


_message_broker_tt_included_test_reply_to = _message_broker_tt_included_params.copy()
_message_broker_tt_included_test_reply_to.update(
    {
        "reply_to": REPLY_TO,
    }
)

_message_broker_tt_forgone_test_reply_to = ["queue_name", "correlation_id", "headers"]


@validate_transaction_metrics(
    "test_pika_produce:test_blocking_connection_reply_to",
    scoped_metrics=_test_blocking_connection_metrics,
    rollup_metrics=_test_blocking_connection_metrics,
    background_task=True,
)
@validate_tt_collector_json(
    message_broker_params=_message_broker_tt_included_test_reply_to,
    message_broker_forgone_params=_message_broker_tt_forgone_test_reply_to,
)
@background_task()
@validate_messagebroker_headers
@cache_pika_headers
def test_blocking_connection_reply_to(producer):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()

        channel.basic_publish(
            exchange="",
            routing_key=QUEUE,
            body="test",
            properties=pika.spec.BasicProperties(reply_to=REPLY_TO),
        )

        # publish has been removed and replaced with basic_publish in later
        # versions of pika
        publish = getattr(channel, "publish", channel.basic_publish)

        publish(
            exchange="",
            routing_key=QUEUE,
            body="test",
            properties=pika.spec.BasicProperties(reply_to=REPLY_TO),
        )


_message_broker_tt_included_test_headers = _message_broker_tt_included_params.copy()
_message_broker_tt_included_test_headers.update(
    {
        "headers": HEADERS.copy(),
    }
)

_message_broker_tt_forgone_test_headers = ["queue_name", "correlation_id", "reply_to"]


@pytest.mark.parametrize("enable_distributed_tracing", [True, False])
def test_blocking_connection_headers(enable_distributed_tracing):

    override_settings = {
        "distributed_tracing.enabled": enable_distributed_tracing,
        "cross_application_tracer.enabled": not enable_distributed_tracing,
    }
    rollup_metrics = list(_test_blocking_connection_metrics)
    if enable_distributed_tracing:
        rollup_metrics += [
            ("DurationByCaller/Unknown/Unknown/Unknown/Unknown/all", 1),
            ("DurationByCaller/Unknown/Unknown/Unknown/Unknown/allOther", 1),
        ]

    @override_application_settings(override_settings)
    @validate_transaction_metrics(
        "test_blocking_connection_headers",
        scoped_metrics=_test_blocking_connection_metrics,
        rollup_metrics=rollup_metrics,
        background_task=True,
    )
    @validate_tt_collector_json(
        message_broker_params=_message_broker_tt_included_test_headers,
        message_broker_forgone_params=_message_broker_tt_forgone_test_headers,
    )
    @background_task(name="test_blocking_connection_headers")
    @validate_messagebroker_headers
    @cache_pika_headers
    def _test():
        with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
            channel = connection.channel()

            channel.basic_publish(
                exchange="",
                routing_key=QUEUE,
                body="test",
                properties=pika.spec.BasicProperties(headers=HEADERS),
            )

            # publish has been removed and replaced with basic_publish in later
            # versions of pika
            publish = getattr(channel, "publish", channel.basic_publish)

            publish(
                exchange="",
                routing_key=QUEUE,
                body="test",
                properties=pika.spec.BasicProperties(headers=HEADERS),
            )

    _test()


@validate_transaction_metrics(
    "test_pika_produce:test_blocking_connection_headers_reuse_properties",
    scoped_metrics=_test_blocking_connection_metrics,
    rollup_metrics=_test_blocking_connection_metrics,
    background_task=True,
)
@validate_tt_collector_json(
    message_broker_params=_message_broker_tt_included_test_headers,
    message_broker_forgone_params=_message_broker_tt_forgone_test_headers,
)
@background_task()
@validate_messagebroker_headers
@cache_pika_headers
def test_blocking_connection_headers_reuse_properties(producer):
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()
        properties = pika.spec.BasicProperties(headers=HEADERS)

        channel.basic_publish(
            exchange="",
            routing_key=QUEUE,
            body="test",
            properties=properties,
        )

        # publish has been removed and replaced with basic_publish in later
        # versions of pika
        publish = getattr(channel, "publish", channel.basic_publish)

        publish(
            exchange="",
            routing_key=QUEUE,
            body="test",
            properties=properties,
        )


_test_blocking_connection_two_exchanges_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/exchange-1", 1),
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/exchange-2", 1),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/exchange-1", None),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/exchange-2", None),
]


@validate_transaction_metrics(
    "test_pika_produce:test_blocking_connection_two_exchanges",
    scoped_metrics=_test_blocking_connection_two_exchanges_metrics,
    rollup_metrics=_test_blocking_connection_two_exchanges_metrics,
    background_task=True,
)
@validate_tt_collector_json(
    message_broker_params=_message_broker_tt_included_params,
    message_broker_forgone_params=_message_broker_tt_forgone_params,
)
@background_task()
@validate_messagebroker_headers
@cache_pika_headers
def test_blocking_connection_two_exchanges():
    with pika.BlockingConnection(pika.ConnectionParameters(DB_SETTINGS["host"])) as connection:
        channel = connection.channel()
        channel.queue_declare(queue=QUEUE)
        channel.exchange_declare(exchange="exchange-1", durable=False, auto_delete=True)
        channel.exchange_declare(exchange="exchange-2", durable=False, auto_delete=True)

        channel.basic_publish(
            exchange="exchange-1",
            routing_key=QUEUE,
            body="test",
        )
        channel.basic_publish(
            exchange="exchange-2",
            routing_key=QUEUE,
            body="test",
        )


_test_select_connection_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/Default", 1),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/Default", None),
]


@validate_transaction_metrics(
    "test_pika_produce:test_select_connection",
    scoped_metrics=_test_select_connection_metrics,
    rollup_metrics=_test_select_connection_metrics,
    background_task=True,
)
@validate_tt_collector_json(
    message_broker_params=_message_broker_tt_included_params,
    message_broker_forgone_params=_message_broker_tt_forgone_params,
)
@background_task()
@validate_messagebroker_headers
@cache_pika_headers
def test_select_connection():
    def on_open(connection):
        connection.channel(on_open_callback=on_channel_open)

    def on_channel_open(channel):
        channel.basic_publish(
            exchange="",
            routing_key=QUEUE,
            body="test",
        )
        connection.close()
        connection.ioloop.stop()

    parameters = pika.ConnectionParameters(DB_SETTINGS["host"])
    connection = pika.SelectConnection(parameters=parameters, on_open_callback=on_open)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        # Start the IOLoop again so Pika can communicate, it will stop on its
        # own when the connection is closed
        connection.ioloop.start()
        raise


_test_tornado_connection_metrics = [
    ("MessageBroker/RabbitMQ/Exchange/Produce/Named/Default", 1),
    ("MessageBroker/RabbitMQ/Exchange/Consume/Named/Default", None),
]


@validate_transaction_metrics(
    "test_pika_produce:test_tornado_connection",
    scoped_metrics=_test_tornado_connection_metrics,
    rollup_metrics=_test_tornado_connection_metrics,
    background_task=True,
)
@validate_tt_collector_json(
    message_broker_params=_message_broker_tt_included_params,
    message_broker_forgone_params=_message_broker_tt_forgone_params,
)
@background_task()
@validate_messagebroker_headers
@cache_pika_headers
def test_tornado_connection():
    from pika.adapters import tornado_connection

    def on_open(connection):
        connection.channel(on_open_callback=on_channel_open)

    def on_channel_open(channel):
        channel.basic_publish(
            exchange="",
            routing_key=QUEUE,
            body="test",
        )
        connection.close()
        connection.ioloop.stop()

    parameters = pika.ConnectionParameters(DB_SETTINGS["host"])
    connection = tornado_connection.TornadoConnection(parameters=parameters, on_open_callback=on_open)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        # Start the IOLoop again so Pika can communicate, it will stop on its
        # own when the connection is closed
        connection.ioloop.start()
        raise



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pytest

from newrelic.api.background_task import background_task
from newrelic.api.time_trace import current_trace
from newrelic.api.transaction import current_transaction
from testing_support.fixtures import reset_core_stats_engine
from testing_support.validators.validate_log_event_count import validate_log_event_count
from testing_support.validators.validate_log_event_count_outside_transaction import validate_log_event_count_outside_transaction
from testing_support.validators.validate_log_events import validate_log_events
from testing_support.validators.validate_log_events_outside_transaction import validate_log_events_outside_transaction


def set_trace_ids():
    txn = current_transaction()
    if txn:
        txn._trace_id = "abcdefgh12345678"
    trace = current_trace()
    if trace:
        trace.guid = "abcdefgh"

def exercise_logging(logger):
    set_trace_ids()

    logger.warning("C")
    logger.error("D")
    logger.critical("E")
    
    assert len(logger.caplog.records) == 3


_common_attributes_service_linking = {"timestamp": None, "hostname": None, "entity.name": "Python Agent Test (logger_loguru)", "entity.guid": None}
_common_attributes_trace_linking = {"span.id": "abcdefgh", "trace.id": "abcdefgh12345678", **_common_attributes_service_linking}

_test_logging_inside_transaction_events = [
    {"message": "C", "level": "WARNING", **_common_attributes_trace_linking},
    {"message": "D", "level": "ERROR", **_common_attributes_trace_linking},
    {"message": "E", "level": "CRITICAL", **_common_attributes_trace_linking},   
]


@reset_core_stats_engine()
def test_logging_inside_transaction(logger):
    @validate_log_events(_test_logging_inside_transaction_events)
    @validate_log_event_count(3)
    @background_task()
    def test():
        exercise_logging(logger)

    test()


_test_logging_outside_transaction_events = [
    {"message": "C", "level": "WARNING", **_common_attributes_service_linking},
    {"message": "D", "level": "ERROR", **_common_attributes_service_linking},
    {"message": "E", "level": "CRITICAL", **_common_attributes_service_linking},   
]

@reset_core_stats_engine()
def test_logging_outside_transaction(logger):
    @validate_log_events_outside_transaction(_test_logging_outside_transaction_events)
    @validate_log_event_count_outside_transaction(3)
    def test():
        exercise_logging(logger)

    test()


@reset_core_stats_engine()
def test_logging_newrelic_logs_not_forwarded(logger):
    @validate_log_event_count(0)
    @background_task()
    def test():
        nr_logger = logging.getLogger("newrelic")
        nr_logger.addHandler(logger.caplog)
        nr_logger.error("A")
        assert len(logger.caplog.records) == 1

    test()


_test_patcher_application_captured_event = {"message": "C-PATCH", "level": "WARNING"}
_test_patcher_application_captured_event.update(_common_attributes_trace_linking)

@reset_core_stats_engine()
def test_patcher_application_captured(logger):
    def patch(record):
        record["message"] += "-PATCH"
        return record

    @validate_log_events([_test_patcher_application_captured_event])
    @validate_log_event_count(1)
    @background_task()
    def test():
        set_trace_ids()
        patch_logger = logger.patch(patch)
        patch_logger.warning("C")

    test()

_test_logger_catch_event = {"level": "ERROR"}  # Message varies wildly, can't be included in test
_test_logger_catch_event.update(_common_attributes_trace_linking)

@reset_core_stats_engine()
def test_logger_catch(logger):
    @validate_log_events([_test_logger_catch_event])
    @validate_log_event_count(1)
    @background_task()
    def test():
        set_trace_ids()

        @logger.catch(reraise=True)
        def throw():
            raise ValueError("Test")

        try:
            with pytest.raises(ValueError):
                throw()
        except ValueError:
            pass
    
    test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import redis
from testing_support.db_settings import redis_settings
from testing_support.fixtures import override_application_settings
from testing_support.util import instance_hostname
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.api.datastore_trace import DatastoreTrace
from newrelic.api.time_trace import current_trace
from newrelic.common.package_version_utils import get_package_version_tuple

DB_SETTINGS = redis_settings()[0]
REDIS_PY_VERSION = get_package_version_tuple("redis")

# Settings

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
}

# Metrics

_base_scoped_metrics = (
    ("Datastore/operation/Redis/scan_iter", 1),
    ("Datastore/operation/Redis/sscan_iter", 1),
    ("Datastore/operation/Redis/zscan_iter", 1),
    ("Datastore/operation/Redis/hscan_iter", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/sadd", 1),
    ("Datastore/operation/Redis/zadd", 1),
    ("Datastore/operation/Redis/hset", 1),
)

_base_rollup_metrics = (
    ("Datastore/all", 8),
    ("Datastore/allOther", 8),
    ("Datastore/Redis/all", 8),
    ("Datastore/Redis/allOther", 8),
    ("Datastore/operation/Redis/scan_iter", 1),
    ("Datastore/operation/Redis/sscan_iter", 1),
    ("Datastore/operation/Redis/zscan_iter", 1),
    ("Datastore/operation/Redis/hscan_iter", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/sadd", 1),
    ("Datastore/operation/Redis/zadd", 1),
    ("Datastore/operation/Redis/hset", 1),
)

_disable_rollup_metrics = list(_base_rollup_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_host = instance_hostname(DB_SETTINGS["host"])
_port = DB_SETTINGS["port"]

_instance_metric_name = "Datastore/instance/Redis/%s/%s" % (_host, _port)

_enable_rollup_metrics.append((_instance_metric_name, 8))

_disable_rollup_metrics.append((_instance_metric_name, None))

# Operations


def exercise_redis(client):
    """
    Exercise client generators by iterating on various methods and ensuring they are
    non-empty, and that traces are started and stopped with the generator.
    """

    # Set existing values
    client.set("scan-key", "value")
    client.sadd("sscan-key", "value")
    client.zadd("zscan-key", {"value": 1})
    client.hset("hscan-key", "field", "value")

    # Check generators
    flag = False
    assert not isinstance(current_trace(), DatastoreTrace)  # Assert no active DatastoreTrace
    for k in client.scan_iter("scan-*"):
        assert k == b"scan-key"
        assert isinstance(current_trace(), DatastoreTrace)  # Assert DatastoreTrace now active
        flag = True
    assert flag

    flag = False
    assert not isinstance(current_trace(), DatastoreTrace)  # Assert no active DatastoreTrace
    for k in client.sscan_iter("sscan-key"):
        assert k == b"value"
        assert isinstance(current_trace(), DatastoreTrace)  # Assert DatastoreTrace now active
        flag = True
    assert flag

    flag = False
    assert not isinstance(current_trace(), DatastoreTrace)  # Assert no active DatastoreTrace
    for k, _ in client.zscan_iter("zscan-key"):
        assert k == b"value"
        assert isinstance(current_trace(), DatastoreTrace)  # Assert DatastoreTrace now active
        flag = True
    assert flag

    flag = False
    assert not isinstance(current_trace(), DatastoreTrace)  # Assert no active DatastoreTrace
    for f, v in client.hscan_iter("hscan-key"):
        assert f == b"field"
        assert v == b"value"
        assert isinstance(current_trace(), DatastoreTrace)  # Assert DatastoreTrace now active
        flag = True
    assert flag


async def exercise_redis_async(client):
    """
    Exercise client generators by iterating on various methods and ensuring they are
    non-empty, and that traces are started and stopped with the generator.
    """

    # Set existing values
    await client.set("scan-key", "value")
    await client.sadd("sscan-key", "value")
    await client.zadd("zscan-key", {"value": 1})
    await client.hset("hscan-key", "field", "value")

    # Check generators
    flag = False
    assert not isinstance(current_trace(), DatastoreTrace)  # Assert no active DatastoreTrace
    async for k in client.scan_iter("scan-*"):
        assert k == b"scan-key"
        assert isinstance(current_trace(), DatastoreTrace)  # Assert DatastoreTrace now active
        flag = True
    assert flag

    flag = False
    assert not isinstance(current_trace(), DatastoreTrace)  # Assert no active DatastoreTrace
    async for k in client.sscan_iter("sscan-key"):
        assert k == b"value"
        assert isinstance(current_trace(), DatastoreTrace)  # Assert DatastoreTrace now active
        flag = True
    assert flag

    flag = False
    assert not isinstance(current_trace(), DatastoreTrace)  # Assert no active DatastoreTrace
    async for k, _ in client.zscan_iter("zscan-key"):
        assert k == b"value"
        assert isinstance(current_trace(), DatastoreTrace)  # Assert DatastoreTrace now active
        flag = True
    assert flag

    flag = False
    assert not isinstance(current_trace(), DatastoreTrace)  # Assert no active DatastoreTrace
    async for f, v in client.hscan_iter("hscan-key"):
        assert f == b"field"
        assert v == b"value"
        assert isinstance(current_trace(), DatastoreTrace)  # Assert DatastoreTrace now active
        flag = True
    assert flag


# Tests


@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_generators:test_strict_redis_generator_enable_instance",
    scoped_metrics=_base_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_strict_redis_generator_enable_instance():
    client = redis.StrictRedis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)
    exercise_redis(client)


@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_generators:test_strict_redis_generator_disable_instance",
    scoped_metrics=_base_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_strict_redis_generator_disable_instance():
    client = redis.StrictRedis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)
    exercise_redis(client)


@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_generators:test_redis_generator_enable_instance",
    scoped_metrics=_base_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_redis_generator_enable_instance():
    client = redis.Redis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)
    exercise_redis(client)


@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_generators:test_redis_generator_disable_instance",
    scoped_metrics=_base_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_redis_generator_disable_instance():
    client = redis.Redis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)
    exercise_redis(client)


@pytest.mark.skipif(REDIS_PY_VERSION < (4, 2), reason="Redis.asyncio was not added until v4.2")
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_generators:test_redis_async_generator_enable_instance",
    scoped_metrics=_base_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_redis_async_generator_enable_instance(loop):
    client = redis.asyncio.Redis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)
    loop.run_until_complete(exercise_redis_async(client))


@pytest.mark.skipif(REDIS_PY_VERSION < (4, 2), reason="Redis.asyncio was not added until v4.2")
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_generators:test_redis_async_generator_disable_instance",
    scoped_metrics=_base_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_redis_async_generator_disable_instance(loop):
    client = redis.asyncio.Redis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)
    loop.run_until_complete(exercise_redis_async(client))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from testing_support.fixtures import catch_background_exceptions
from newrelic.common.object_wrapper import transient_function_wrapper, function_wrapper


def validate_dimensional_metrics_outside_transaction(dimensional_metrics=None):
    dimensional_metrics = dimensional_metrics or []

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):

        record_dimensional_metric_called = []
        recorded_metrics = [None]

        @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_dimensional_metric")
        @catch_background_exceptions
        def _validate_dimensional_metrics_outside_transaction(wrapped, instance, args, kwargs):
            record_dimensional_metric_called.append(True)
            try:
                result = wrapped(*args, **kwargs)
            except:
                raise
            else:
                metrics = instance.dimensional_stats_table.metrics()
                # Record a copy of the metric value so that the values aren't
                # merged in the future
                _metrics = {}
                for k, v in metrics:
                    _metrics[k] = copy.copy(v)
                recorded_metrics[0] = _metrics

            return result

        def _validate(metrics, name, tags, count):
            key = (name, tags)
            # Dimensional metric lookup
            metric_container = metrics.get(name, {})
            metric = metric_container.get(tags)

            def _metrics_table():
                out = [""]
                out.append("Expected: {0}: {1}".format(key, count))
                for metric_key, metric_container in metrics.items():
                    if isinstance(metric_container, dict):
                        for metric_tags, metric_value in metric_container.items():
                            out.append("{0}: {1}".format((metric_key, metric_tags), metric_value[0]))
                    else:
                        out.append("{0}: {1}".format(metric_key, metric_container[0]))
                return "\n".join(out)

            def _metric_details():
                return "metric=%r, count=%r" % (key, metric.call_count)

            if count is not None:
                assert metric is not None, _metrics_table()
                if count == "present":
                    assert metric.call_count > 0, _metric_details()
                else:
                    assert metric.call_count == count, _metric_details()

                assert metric.total_call_time >= 0, (key, metric)
                assert metric.total_exclusive_call_time >= 0, (key, metric)
                assert metric.min_call_time >= 0, (key, metric)
                assert metric.sum_of_squares >= 0, (key, metric)

            else:
                assert metric is None, _metrics_table()

        _new_wrapper = _validate_dimensional_metrics_outside_transaction(wrapped)
        val = _new_wrapper(*args, **kwargs)
        assert record_dimensional_metric_called
        metrics = recorded_metrics[0]

        record_dimensional_metric_called[:] = []
        recorded_metrics[:] = []

        for dimensional_metric, dimensional_tags, count in dimensional_metrics:
            if isinstance(dimensional_tags, dict):
                dimensional_tags = frozenset(dimensional_tags.items())
            _validate(metrics, dimensional_metric, dimensional_tags, count)

        return val

    return _validate_wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import json

from framework_ariadne._target_schema_async import (
    target_asgi_application as target_asgi_application_async,
)
from framework_ariadne._target_schema_async import target_schema as target_schema_async
from framework_ariadne._target_schema_sync import (
    target_asgi_application as target_asgi_application_sync,
)
from framework_ariadne._target_schema_sync import target_schema as target_schema_sync
from framework_ariadne._target_schema_sync import (
    target_wsgi_application as target_wsgi_application_sync,
)
from framework_ariadne._target_schema_sync import ariadne_version_tuple
from graphql import MiddlewareManager



def check_response(query, success, response):
    if isinstance(query, str) and "error" not in query:
        assert success and "errors" not in response, response
        assert response.get("data", None), response
    else:
        assert "errors" in response, response


def run_sync(schema):
    def _run_sync(query, middleware=None):
        from ariadne import graphql_sync

        if ariadne_version_tuple < (0, 18):
            if middleware:
                middleware = MiddlewareManager(*middleware)

        success, response = graphql_sync(schema, {"query": query}, middleware=middleware)
        check_response(query, success, response)

        return response.get("data", {})

    return _run_sync


def run_async(schema):
    def _run_async(query, middleware=None):
        from ariadne import graphql

        #Later versions of ariadne directly accept a list of middleware while older versions require the MiddlewareManager
        if ariadne_version_tuple < (0, 18):
            if middleware:
                middleware = MiddlewareManager(*middleware)

        loop = asyncio.get_event_loop()
        success, response = loop.run_until_complete(graphql(schema, {"query": query}, middleware=middleware))
        check_response(query, success, response)

        return response.get("data", {})

    return _run_async


def run_wsgi(app):
    def _run_asgi(query, middleware=None):
        if not isinstance(query, str) or "error" in query:
            expect_errors = True
        else:
            expect_errors = False

        app.app.middleware = middleware

        response = app.post(
            "/", json.dumps({"query": query}), headers={"Content-Type": "application/json"}, expect_errors=expect_errors
        )

        body = json.loads(response.body.decode("utf-8"))
        if expect_errors:
            assert body["errors"]
        else:
            assert "errors" not in body or not body["errors"]

        return body.get("data", {})

    return _run_asgi


def run_asgi(app):
    def _run_asgi(query, middleware=None):
        if ariadne_version_tuple < (0, 16):
            app.asgi_application.middleware = middleware

        #In ariadne v0.16.0, the middleware attribute was removed from the GraphQL class in favor of the http_handler
        elif ariadne_version_tuple >= (0, 16):
            app.asgi_application.http_handler.middleware = middleware

        response = app.make_request(
            "POST", "/", body=json.dumps({"query": query}), headers={"Content-Type": "application/json"}
        )
        body = json.loads(response.body.decode("utf-8"))

        if not isinstance(query, str) or "error" in query:
            try:
                assert response.status != 200
            except AssertionError:
                assert body["errors"]
        else:
            assert response.status == 200
            assert "errors" not in body or not body["errors"]

        return body.get("data", {})

    return _run_asgi


target_application = {
    "sync-sync": run_sync(target_schema_sync),
    "async-sync": run_async(target_schema_sync),
    "async-async": run_async(target_schema_async),
    "wsgi-sync": run_wsgi(target_wsgi_application_sync),
    "asgi-sync": run_asgi(target_asgi_application_sync),
    "asgi-async": run_asgi(target_asgi_application_async),
}



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import six

from newrelic.common.object_wrapper import function_wrapper, transient_function_wrapper

try:
    from newrelic.core.infinite_tracing_pb2 import AttributeValue, Span
except:
    AttributeValue = None
    Span = None


def validate_span_events(
    count=1,
    exact_intrinsics={},
    expected_intrinsics=[],
    unexpected_intrinsics=[],
    exact_agents={},
    expected_agents=[],
    unexpected_agents=[],
    exact_users={},
    expected_users=[],
    unexpected_users=[],
    index=-1,
):

    # Used for validating a single span event.
    #
    # Since each transaction could produce multiple span events, assert that at
    # least `count` number of those span events meet the criteria, if not,
    # raises an AssertionError.
    #
    # Use this validator once per distinct span event expected.

    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):

        record_transaction_called = []
        recorded_span_events = []

        @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
        def capture_span_events(wrapped, instance, args, kwargs):
            events = []

            @transient_function_wrapper("newrelic.common.streaming_utils", "StreamBuffer.put")
            def stream_capture(wrapped, instance, args, kwargs):
                event = args[0]
                events.append(event)
                return wrapped(*args, **kwargs)

            record_transaction_called.append(True)
            try:
                result = stream_capture(wrapped)(*args, **kwargs)
            except:
                raise
            else:
                if not instance.settings.infinite_tracing.enabled:
                    events = [event for priority, seen_at, event in instance.span_events.pq]

                recorded_span_events.append(events)

            return result

        _new_wrapper = capture_span_events(wrapped)
        val = _new_wrapper(*args, **kwargs)
        assert record_transaction_called
        captured_events = recorded_span_events.pop(index)

        mismatches = []
        matching_span_events = 0
        for captured_event in captured_events:
            if Span and isinstance(captured_event, Span):
                intrinsics = captured_event.intrinsics
                user_attrs = captured_event.user_attributes
                agent_attrs = captured_event.agent_attributes
            else:
                intrinsics, user_attrs, agent_attrs = captured_event

            _check_span_intrinsics(intrinsics)

            intrinsics_ok = _check_span_attributes(
                intrinsics, exact_intrinsics, expected_intrinsics, unexpected_intrinsics, mismatches
            )
            agent_attr_ok = _check_span_attributes(
                agent_attrs, exact_agents, expected_agents, unexpected_agents, mismatches
            )
            user_attr_ok = _check_span_attributes(user_attrs, exact_users, expected_users, unexpected_users, mismatches)

            if intrinsics_ok and agent_attr_ok and user_attr_ok:
                matching_span_events += 1

        def _span_details():
            details = [
                "matching_span_events=%d" % matching_span_events,
                "count=%d" % count,
                "mismatches=%s" % mismatches,
                "captured_events=%s" % captured_events,
            ]

            return "\n".join(details)

        assert matching_span_events == count, _span_details()

        return val

    return _validate_wrapper


def check_value_equals(dictionary, key, expected_value):
    value = dictionary.get(key)
    if AttributeValue and isinstance(value, AttributeValue):
        for descriptor, val in value.ListFields():
            if val != expected_value:
                return False
        return True
    else:
        return value == expected_value


def assert_isinstance(value, expected_type):
    if AttributeValue and isinstance(value, AttributeValue):
        if expected_type is six.string_types:
            assert value.HasField("string_value")
        elif expected_type is float:
            assert value.HasField("double_value")
        elif expected_type is int:
            assert value.HasField("int_value")
        else:
            assert False
    else:
        assert isinstance(value, expected_type)


def _check_span_attributes(attrs, exact, expected, unexpected, mismatches):
    for expected_attribute in expected:
        if expected_attribute not in attrs:
            break
    else:
        for unexpected_attribute in unexpected:
            if unexpected_attribute in attrs:
                break
        else:
            for key, value in exact.items():
                if not check_value_equals(attrs, key, value):
                    mismatches.append("key: %s, value:<%s><%s>" % (key, value, attrs.get(key)))
                    break
            else:
                return True

    return False


def _check_span_intrinsics(intrinsics):
    assert check_value_equals(intrinsics, "type", "Span")
    assert_isinstance(intrinsics["traceId"], six.string_types)
    assert_isinstance(intrinsics["guid"], six.string_types)
    if "parentId" in intrinsics:
        assert_isinstance(intrinsics["parentId"], six.string_types)
    assert_isinstance(intrinsics["transactionId"], six.string_types)
    intrinsics["sampled"] is True
    assert_isinstance(intrinsics["priority"], float)
    assert_isinstance(intrinsics["timestamp"], int)
    ts = intrinsics["timestamp"]
    if AttributeValue and isinstance(ts, AttributeValue):
        ts = ts.double_value
    assert ts <= int(time.time() * 1000)
    assert_isinstance(intrinsics["duration"], float)
    assert_isinstance(intrinsics["name"], six.string_types)
    assert_isinstance(intrinsics["category"], six.string_types)
    if "nr.entryPoint" in intrinsics:
        assert check_value_equals(intrinsics, "nr.entryPoint", True)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from newrelic.core.config import global_settings
from testing_support.fixtures import (
        override_ignore_status_codes,
        override_generic_settings)
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors

SETTINGS = global_settings()


def test_basic(app):
    _test_basic_metrics = (
        ('Function/' + app.name_prefix + '.__call__', 1),
        ('Function/_target_application:Index.on_get', 1),
    )

    @validate_code_level_metrics("_target_application.Index", "on_get")
    @validate_transaction_metrics('_target_application:Index.on_get',
            scoped_metrics=_test_basic_metrics,
            rollup_metrics=_test_basic_metrics)
    def _test():
        response = app.get('/', status=200)
        response.mustcontain('ok')

    _test()


@override_ignore_status_codes([404])
@validate_transaction_errors(errors=[])
def test_ignored_status_code(app):

    @validate_transaction_metrics(app.name_prefix + '._handle_exception')
    def _test():
        app.get('/foobar', status=404)

    _test()


@override_ignore_status_codes([])
def test_error_recorded(app):

    @validate_transaction_errors(errors=[app.not_found_error])
    @validate_transaction_metrics(app.name_prefix + '._handle_exception')
    def _test():
        app.get('/foobar', status=404)

    _test()


# This test verifies that we don't actually break anything if somebody puts
# garbage into the status code
@validate_transaction_metrics('_target_application:BadResponse.on_get')
@validate_code_level_metrics("_target_application.BadResponse", "on_get")
@validate_transaction_errors(errors=['_target_application:BadGetRequest'])
def test_bad_response_error(app):
    # Disable linting since this should actually be an invalid response
    # (incorrect media type int)
    lint = app.lint
    app.lint = False
    try:
        app.get('/bad_response', status=200)
    finally:
        app.lint = lint


@validate_transaction_metrics('_target_application:BadResponse.on_put')
@validate_code_level_metrics("_target_application.BadResponse", "on_put")
@validate_transaction_errors(errors=['_target_application:BadPutRequest'])
def test_unhandled_exception(app):
    from falcon import __version__ as falcon_version

    # Falcon v3 and above will not raise an uncaught exception
    if int(falcon_version.split('.', 1)[0]) >= 3:
        app.put('/bad_response', status=500, expect_errors=True)
    else:
        with pytest.raises(app.BadPutRequest):
            app.put('/bad_response')


@override_generic_settings(SETTINGS, {
    'enabled': False,
})
def test_nr_disabled_ok(app):
    response = app.get('/', status=200)
    response.mustcontain('ok')


@override_generic_settings(SETTINGS, {
    'enabled': False,
})
def test_nr_disabled_error(app):
    app.get('/foobar', status=404)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

import pytest

from testing_support.fixtures import collector_agent_registration_fixture, collector_available_fixture  # noqa: F401; pylint: disable=W0611
from testing_support.mock_external_grpc_server import MockExternalgRPCServer

from newrelic.common.streaming_utils import StreamBuffer

CONDITION_CLS = type(threading.Condition())


_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
    "debug.log_autorum_middleware": True,
    "agent_limits.errors_per_harvest": 100,
    "distributed_tracing.enabled": True,
    "infinite_tracing.trace_observer_host": "nr-internal.aws-us-east-2.tracing.staging-edge.nr-data.net",
    "infinite_tracing.compression": True,
    "debug.connect_span_stream_in_developer_mode": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (agent_streaming)", default_settings=_default_settings
)


@pytest.fixture(scope="module")
def grpc_app_server():
    with MockExternalgRPCServer() as server:
        yield server, server.port


@pytest.fixture(scope="module")
def mock_grpc_server(grpc_app_server):
    from _test_handler import HANDLERS

    server, port = grpc_app_server
    server.add_generic_rpc_handlers(HANDLERS)
    return port


class SetEventOnWait(CONDITION_CLS):
    def __init__(self, event, *args, **kwargs):
        super(SetEventOnWait, self).__init__(*args, **kwargs)
        self._event = event

    def wait(self, *args, **kwargs):
        self._event.set()
        return super(SetEventOnWait, self).wait(*args, **kwargs)


@pytest.fixture(scope="function")
def buffer_empty_event(monkeypatch):
    event = threading.Event()

    @staticmethod
    def condition(*args, **kwargs):
        return SetEventOnWait(event, *args, **kwargs)

    monkeypatch.setattr(StreamBuffer, "condition", condition)
    return event


@pytest.fixture(scope="session", params=[pytest.param(True, id="batching"), pytest.param(False, id="nonbatching")])
def batching(request):
    return request.param


@pytest.fixture(scope="function")
def spans_received():
    from _test_handler import SPANS_RECEIVED

    SPANS_RECEIVED.clear()
    return SPANS_RECEIVED


@pytest.fixture(scope="function")
def span_batches_received():
    from _test_handler import SPAN_BATCHES_RECEIVED

    SPAN_BATCHES_RECEIVED.clear()
    return SPAN_BATCHES_RECEIVED


@pytest.fixture(scope="function")
def spans_processed_event():
    from _test_handler import SPANS_PROCESSED_EVENT

    SPANS_PROCESSED_EVENT.clear()
    return SPANS_PROCESSED_EVENT
